/* A list of names.
 * A list is stored as a sorted array, with binary search --
 * maybe not the most efficient way, but it's the most optimal
 * from the memory requiriments point of view, simple to
 * handle and sufficient enouth (imho)
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rbldnsd.h"
#include "dns.h"
#include "mempool.h"
#include "sstr.h"

struct zonedata {
  unsigned nent;
  unsigned char **ent;
  char *txt;
  struct in_addr a;
  struct mempool mp;
};

static struct zonedata *freezone(struct zonedata *z) {
  if (z) {
    mp_free(&z->mp);
    if (z->txt) free(z->txt);
    if (z->ent) free(z->ent);
    free(z);
  }
  return NULL;
}

int nameset_query(const struct zonedata *const z,
	    struct dnspacket *p,
	    const unsigned char *const query, unsigned qflg)
{
   int a = 0, b = z->nent - 1, m, r;
   unsigned char **e = z->ent;
   while(a <= b) {
     m = (a + b) >> 1;
     r = strcmp(query, e[m]);
     if (!r) {
       if (qflg & QUERY_A) addrec_a(p, z->a);
       if (qflg & QUERY_TXT && z->txt) {
         char dnname[DNS_MAXDOMAIN];
         addrec_txt(p, z->txt,
                    dns_dntop(e[m], dnname, sizeof(dnname)) ? dnname : NULL);
       }
       return 1;
     }
     else if (r < 0) b = m - 1;
     else a = m + 1;
   }
   return 0;
}

struct zonedata *nameset_load(const char *name, struct zonedata *z) {
  int lineno = 0;
  int l;
  unsigned char **ent = NULL;
  unsigned nent = 0, nalc = 0;
  struct in_addr a;
  char *p;
  unsigned char dn[DNS_MAXDN+1];
  FILE *f;
  char buf[128];
  freezone(z);
  if ((f = fopen(name, "r")) == NULL) {
    log(LOG_ERR, "unable to open zone file `%s': %s", name, strerror(errno));
    return NULL;
  }
  if ((z = (struct zonedata *)emalloc(sizeof(*z))) == NULL) {
    fclose(f);
    return NULL;
  }
  mp_init(&z->mp);
  z->ent = NULL;
  z->txt = NULL;
  z->a = default_aaddr;

  while(fgets(buf, sizeof(buf), f)) {
    ++lineno;
    l = strlen(buf);
    if (l && buf[l-1] == '\n') buf[--l] = '\0';
    if (!l || buf[0] == '#') continue;
    if (lineno == 1 && buf[0] == ':') {
      if (!addrtxt(buf, &a, &p))
        log(LOG_WARNING, "zone %s line 1: invalid default entry", name);
      else {
        z->a = a;
	if (!*p)
          z->txt = NULL;
	else if ((z->txt = esstrdup(p)) == NULL) {
	  fclose(f);
	  return freezone(z);
	}
      }
      continue;
    }
    for(p = buf; *p; ++p)
      if (*p == ' ' || *p == '\t' || *p == '#') break;
    *p++ = '\0';
    if (!dns_ptodn(buf, dn, sizeof(dn))) {
      log(LOG_WARNING, "zone %s, line %d: invalid domain name %s",
          name, lineno, buf);
      continue;
    }
    dns_dntol(dn, dn, sizeof(dn));
    if (nent == nalc) {
      nalc = nalc ? nalc * 2 : 512;
      z->ent = ent = (unsigned char **)erealloc(ent, nalc * sizeof(*ent));
      if (!ent) {
	fclose(f);
	return freezone(z);
      }
    }
    if ((ent[nent] = mp_estrdup(&z->mp, dn)) == NULL) {
      fclose(f);
      return freezone(z);
    }
    ++nent;
  }
  fclose(f);
  z->ent = (unsigned char **)realloc(ent, nent * sizeof(*ent));
  z->nent = nent;
  qsort(ent, nent, sizeof(*ent), (int(*)(const void*,const void*))strcmp);
  return z;
}
